import logging
import os
import unittest
from finbourne_sdk_utils import logger
from finbourne_sdk_utils.cocoon.instruments import prepare_key
from parameterized import parameterized


class CocoonUtilitiesTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    @parameterized.expand(
        [
            [
                "Standard full key",
                "Instrument/default/Figi",
                True,
                "Instrument/default/Figi",
            ],
            [
                "Full key specified as short key",
                "Figi",
                True,
                "Instrument/default/Figi",
            ],
            [
                "Full key specified as property",
                "Instrument/PB/Isin",
                True,
                "Instrument/PB/Isin",
            ],
            ["Standard short key", "Isin", False, "Isin"],
            [
                "Short key specifed as long key",
                "Instrument/default/Isin",
                False,
                "Isin",
            ],
        ]
    )
    def test_create_identifiers_prepare_key(
        self, _, identifier_lusid, full_key_format, expected_outcome
    ) -> None:
        """
        Tests that key preparation for identifiers works as expected

        :param _: The name of the test
        :param str identifier_lusid: The LUSID identifier
        :param bool full_key_format: The full key format
        :param str expected_outcome: The expected output key

        :return: None
        """

        output_key = prepare_key(
            identifier_lusid=identifier_lusid, full_key_format=full_key_format
        )

        self.assertEqual(output_key, expected_outcome)



import yaml
from typing import Callable
from lusid import models

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

# Common abbreviations
rev_abbrev = {
    "FIFO": "TransactionConfiguration/default/TaxLotSelectionMethod=FirstInFirstOut",
    "LIFO": "TransactionConfiguration/default/TaxLotSelectionMethod=LastInFirstOut",
    "HCF": "TransactionConfiguration/default/TaxLotSelectionMethod=HighestCostFirst",
    "LCF": "TransactionConfiguration/default/TaxLotSelectionMethod=LowestCostFirst",
    "AC": "TransactionConfiguration/default/TaxLotSelectionMethod=AverageCost",
    "LL": "LongLonger",
    "SL": "ShortLonger",
    "LS": "LongShorter",
    "SS": "ShortShorter",
    "S1": "Side1",
    "S2": "Side2",
    "STOCK": "Settlement, Traded",
    "CASH": "Commitment, CashSettlement",
    "FWD": "ForwardFx, CashSettlement",
    "RCV": "Receivable, CashSettlement",
}

# Reverse key map for lookups
fwd_abbrev = {v: k for k, v in rev_abbrev.items()}


# Methods to map between abbreviations
def abbrev(v):
    return fwd_abbrev.get(v, v)


def unabbrev(v):
    return rev_abbrev.get(v, v)


# Dumper class with the custom constructor/resolvers
class CustomDumper(yaml.Dumper):
    pass


# Loader class with the custom constructor for querying
# This is required because the input message uses different classes
class QueryLoader(yaml.Loader):
    pass


# Loader class with the custom constructor for updates
# This is required because the input message uses different classes
class UpdateLoader(yaml.Loader):
    pass


class TxnConfigYaml:
    """
    This class holds all the configuration and methods for working with the TransactionTypeConfiguration

    Each object in the TransactionTypeConfiguration has a representation method which converts a LUSID model to YAML
    with the suffix "_rep" and a constructor method which converts YAML to a LUSID model with the suffix "_con".
    """

    def __init__(
        self,
        models,
        UpdateLoader: yaml.Loader = UpdateLoader,
        QueryLoader: yaml.Loader = QueryLoader,
        CustomDumper: yaml.Dumper = CustomDumper,
    ):
        """
        The initialisation here will set up the custom constructors and representers. It will add these to the the
        CustomDumper, QueryLoader and UpdateLoader.

        :param models: The lusid.models
        :param yaml.Loader UpdateLoader: The loader to use for updating (writing to LUSID)
        :param yaml.Loader QueryLoader: The loader to use for querying (reading from LUSID)
        :param yaml.Dumper CustomDumper: The dumper used for converting LUSID models to YAML
        """
        self.YAMLConverter(
            "!Txn",
            self.config_rep,
            self.TransactionSetConfigurationDataNoLinks,
            self.config_con,
            self.TransactionSetConfigurationDataNoLinks,
            models.TransactionSetConfigurationDataRequest,
        )

        self.YAMLConverter(
            "!TxnConfig",
            self.trans_config_rep,
            models.TransactionConfigurationData,
            self.trans_config_con,
            models.TransactionConfigurationData,
            models.TransactionConfigurationDataRequest,
        )

        self.YAMLConverter(
            "!Mvmt",
            self.mvmt_rep,
            models.TransactionConfigurationMovementData,
            self.mvmt_con,
            models.TransactionConfigurationMovementData,
            models.TransactionConfigurationMovementDataRequest,
        )

        self.YAMLConverter(
            "!Map",
            self.pm_rep,
            models.TransactionPropertyMapping,
            self.pm_con,
            models.TransactionPropertyMapping,
            models.TransactionPropertyMappingRequest,
        )

        self.YAMLConverter(
            "!Alias",
            self.alias_rep,
            models.TransactionConfigurationTypeAlias,
            self.alias_con,
            models.TransactionConfigurationTypeAlias,
            models.TransactionConfigurationTypeAlias,
        )

        self.YAMLConverter(
            "!SideConfig",
            self.side_config_rep,
            models.SideConfigurationData,
            self.side_config_con,
            models.SideConfigurationData,
            models.SideConfigurationDataRequest,
        )

        self.YAMLConverter(
            "!Prop",
            self.pp_rep,
            models.PerpetualProperty,
            self.pp_con,
            models.PerpetualProperty,
            models.PerpetualProperty,
        )

    class YAMLConverter:
        """
        This class handles the configuration for a single element of the model e.g. aliases on a
        TransactionTypeConfiguration.
        """

        def __init__(
            self,
            tag: str,
            representation_function: Callable,
            representation_model,
            constructor_function: Callable,
            constructor_read_model,
            constructor_update_model,
            representation_dumper: yaml.Dumper = CustomDumper,
            constructor_read_loader: yaml.Loader = QueryLoader,
            constructor_update_loader: yaml.Loader = UpdateLoader,
        ):
            """
            The initialisation function creates the representation and constructor functions from the provided
            base callable functions and registers them with the dumper and loaders.

            :param str tag: The tag to use for this node
            :param Callable representation_function: The base representation function to use when converting to YAML
            :param lusid.models representation_model: The model to convert from when representing as YAML
            :param constructor_function: The base constructor function to use when converting from YAML
            :param lusid.models constructor_read_model: The model to use for reading from LUSID
            :param lusid.models constructor_update_model: The model to use for sending to LUSID
            (usually has "Request" added to the name)
            :param yaml.Dumper representation_dumper: The dumper to use for converting to YAML
            :param yaml.Loader constructor_read_loader: The loader to use for converting from YAML when reading from LUSID
            :param yaml.Loader constructor_update_loader: The loader to use for converting from YAML when writing to LUSID
            """

            # Maps the types of the element to the appropriate dumper method
            type_mapping = {
                "<class 'dict'>": "represent_mapping",
                "<class 'str'>": "represent_scalar",
                "<class 'list'>": "represent_sequence",
            }

            def representation_function_dumper(
                dumper: yaml.Dumper, data, repr: Callable = representation_function
            ) -> Callable:
                """
                This function is used by the dumper to turn a LUSID model object into YAML

                :param yaml.Dumper dumper: The dumper to use
                :param data: The data
                :param Callable repr: The base representation function

                :return: Callable
                """
                # Get the payload from the representation function
                payload = repr(data)
                # Get the appropriate dumper function to use
                representation_type = type_mapping[str((type(payload)))]

                # Construct and return the appropriate dumper representation function
                if representation_type == "represent_sequence":
                    return getattr(dumper, representation_type)(
                        tag, payload, flow_style=True
                    )
                else:
                    return getattr(dumper, representation_type)(tag, payload)

            # Add the dumper function
            yaml.add_representer(
                data_type=representation_model,
                representer=representation_function_dumper,
                Dumper=representation_dumper,
            )

            def constructor_function_read(
                loader: yaml.Loader, node, con: Callable = constructor_function
            ):
                """
                This function is used to construct a LUSID model from YAML when reading from LUSID

                :param yaml.Loader loader: The YAML loader to use
                :param node: The YAML node to construct the model from
                :param con: The base constructor function

                :return: lusid.models: The populated LUSID model
                """
                return constructor_read_model(*con(loader, node))

            def constructor_function_update(loader, node, con=constructor_function):
                """
                This function is used to construct a LUSID model from YAML when writing to LUSID. The only difference
                from the "constructor_function_read" is that it has a different model.

                :param yaml.Loader loader: The YAML loader to use
                :param node: The YAML node to construct the model from
                :param con: The base constructor function

                :return: lusid.models: The populated LUSID model
                """
                return constructor_update_model(*con(loader, node))

            # Add the YAML to LUSID model constructor for reading from LUSID
            yaml.add_constructor(
                tag=tag,
                constructor=constructor_function_read,
                Loader=constructor_read_loader,
            )

            # Add the YAML to LUSID model constructor for writing to LUSID
            yaml.add_constructor(
                tag=tag,
                constructor=constructor_function_update,
                Loader=constructor_update_loader,
            )

    @staticmethod
    def find_by_tag(node: yaml.Node, tag: str):
        """
        Utility function which searches the node until the specified tag is first found returning the discovered node

        :param yaml.Node node: The node to search
        :param str tag: The tag to search for in the node

        :return: yaml.Node: The first discovered node which matches the tag
        """
        for i in node.value:
            if i[0].value == tag:
                return i[1]

    @staticmethod
    def prepare_properties(properties_node_list: list) -> dict:
        """
        Utility which prepares a single property from a properties node so that it can be sent to LUSID as a dictionary

        :param list properties_node_list: The properties node to prepare

        :return dict(str, models.PerpetualProperty) properties: The property in a dictionary
        """
        if len(properties_node_list) > 0:
            single_property = {n.key: n for n in properties_node_list}
        else:
            single_property = {}

        return single_property

    # Set up the YAML converter for the TransactionSetConfigurationData (top level object)
    @staticmethod
    def config_rep(data):
        return {
            "transactionConfigRequests": data.transaction_configs,
            "sideConfigRequests": data.side_definitions,
        }

    @classmethod
    def config_con(cls, loader, node):
        txn = loader.construct_sequence(
            cls.find_by_tag(node, "transactionConfigRequests")
        )
        side = loader.construct_sequence(cls.find_by_tag(node, "sideConfigRequests"))
        return txn, side

    # Set up the YAML converter for the TransactionConfigurationData (transaction_configs)
    @staticmethod
    def trans_config_rep(data):
        return {
            "aliases": data.aliases,
            "movements": data.movements,
            "properties": list(data.properties.values()),
        }

    @classmethod
    def trans_config_con(cls, loader, node):
        a = loader.construct_sequence(cls.find_by_tag(node, "aliases"))
        m = loader.construct_sequence(cls.find_by_tag(node, "movements"))
        p = cls.prepare_properties(
            loader.construct_sequence(cls.find_by_tag(node, "properties"))
        )
        return a, m, p

    # Set up the YAML converter for the TransactionConfigurationMovementData (movements)
    @staticmethod
    def mvmt_rep(data):
        return [
            [abbrev(data.side), data.direction, abbrev(data.movement_types)],
            list(data.properties.values()),
            data.mappings,
        ] + ([data.name] if data.name else [])

    @classmethod
    def mvmt_con(cls, loader, node):
        s = loader.construct_sequence(node.value[0])
        properties = cls.prepare_properties(loader.construct_sequence(node.value[1]))
        side = unabbrev(s[0])
        direction = s[1]
        movement_types = unabbrev(s[2])
        mappings = loader.construct_sequence(node.value[2])

        name = None
        if len(node.value) > 3:
            name = node.value[3].value
            if name == "null":
                name = None

        return movement_types, side, direction, properties, mappings, name

    # Set up the YAML converter for the TransactionPropertyMapping (mappings)
    @staticmethod
    def pm_rep(data):
        if data.map_from is not None:
            return "{}=F:{}".format(data.property_key, data.map_from)
        else:
            return "{}=S:{}".format(data.property_key, data.set_to)

    @staticmethod
    def pm_con(loader, node):
        s = node.value.split("=")

        property_key = s[0]

        if s[1].startswith("F:"):
            map_from = s[1][2:]
            set_to = None
        else:
            set_to = s[1][2:]
            map_from = None

        return property_key, map_from, set_to

    # Set up the YAML converter for the TransactionConfigurationTypeAlias (aliases)
    @staticmethod
    def alias_rep(data):
        return [
            data.type,
            data.description,
            data.transaction_group,
            data.source,
            data.transaction_class,
            abbrev(data.transaction_roles),
        ] + (["default"] if data.is_default else [])

    @staticmethod
    def alias_con(loader, node):
        s = loader.construct_sequence(node)
        # If old yaml files are used with only 5 params set for txn type alias (SENG-41)
        if len(s) == 5:
            return (
                s[0],
                s[1],
                s[3],
                s[3],
                s[2],
                unabbrev(s[4]),
                (len(s) > 5 and s[5] == "default"),
            )
        # If newer yaml files are used with additional params set for txn type alias (SENG-41)
        else:
            return (
                s[0],
                s[1],
                s[4],
                s[3],
                s[2],
                unabbrev(s[5]),
                (len(s) > 6 and s[6] == "default"),
            )

    # Set up the YAML converter for the SideConfigurationDataRequest (side)
    @staticmethod
    def side_config_rep(data):
        return [
            data.side,
            data.security,
            data.currency,
            data.rate,
            data.units,
            data.amount,
        ]

    @staticmethod
    def side_config_con(loader, node):
        s = loader.construct_sequence(node)
        side = s[0]
        security = s[1]
        currency = s[2]
        rate = s[3]
        units = s[4]
        amount = s[5]
        return side, security, currency, rate, units, amount

    # Set up the YAML converter for PerpetualProperties (properties)
    @staticmethod
    def pp_rep(data):
        return abbrev("{}={}".format(data.key, data.value.label_value))

    @staticmethod
    def pp_con(loader, node):
        s = unabbrev(node.value).split("=")
        return (s[0], models.property_value.PropertyValue(s[1]))

    # END OF THE INITIALISER ############################################

    # Write the yaml file with validation that the custom view doesn't drop any
    # important data
    class TransactionSetConfigurationDataNoLinks:
        """
        This class is used to replace TransactionSetConfigurationData so that it has no extra information e.g. links
        which get converted to YAML.
        """

        def __init__(self, transaction_configs, side_definitions):
            self.transaction_configs = transaction_configs
            self.side_definitions = side_definitions

    def dump(self, obj, filename, raw=False):

        if raw:
            with open(filename, "w") as stream:
                yaml.dump(obj, stream)
        else:
            orig = yaml.dump(obj, width=500)
            cust = yaml.dump(obj, Dumper=CustomDumper, width=500)
            copy = yaml.dump(yaml.load(cust, Loader=QueryLoader), width=500)

            if orig != copy:
                print("DIFFS - writing to orig/copy")
                with open("orig", "w") as stream:
                    stream.write(orig)
                with open("copy", "w") as stream:
                    stream.write(copy)
            with open(filename, "w") as stream:
                stream.write(cust)

    # Get the YAML using the custom view
    def get_yaml(self, obj):
        return yaml.dump(obj, Dumper=CustomDumper)

    # Read a YAML file
    def load(self, filename):
        with open(filename, "r") as stream:
            return yaml.load(stream, Loader=QueryLoader)

    # Read a YAML file and convert to the UPDATE request message format
    def load_update(self, filename):
        with open(filename, "r") as stream:
            return yaml.load(stream, Loader=UpdateLoader)

    # Read a YAML string and convert to the UPDATE request message format
    def load_update_str(self, s):
        return yaml.load(s, Loader=UpdateLoader)



# Make backwards compatibility adjustments
# When reading old files
def convert(frame_type, df):

    # Security is now known as instrument_id
    if "security_uid" in df.columns.values:
        df.rename(columns={"security_uid": "instrument_uid"}, inplace=True)

    # Some older dividend files may not have the pay date column
    # In this case we duplicate the record date
    if frame_type == "div":
        if "payment_date" not in df.columns.values:
            df["payment_date"] = df["record_date"]

    # Some older pricing files call the instr column 'sec'
    if frame_type == "prc":
        if "instr" not in df.columns.values:
            df["sec"] = df["instr"]

    return df



import unittest
from pathlib import Path

import numpy as np
import pandas as pd
from parameterized import parameterized

import finbourne_sdk_utils.cocoon as cocoon
from finbourne_sdk_utils import logger
from  .mock_api_factory import MockApiFactory
import lusid


class CocoonInvalidPropertiesTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.api_factory = MockApiFactory( )
        cls.logger = logger.LusidLogger("debug")

    @parameterized.expand(
        [
            (
                "Invalid numeric type",
                pd.DataFrame([{"a": 2}]).astype(np.short),
                "The following columns in the data_frame have not been mapped to LUSID data types: {'a': 'int16'}",
            )
        ]
    )
    def test_create_property_definitions_from_file(
        self, _, data_frame, expected_message
    ) -> None:
        with self.assertRaises(TypeError) as context:
            cocoon.properties.create_property_definitions_from_file(
                api_factory=self.api_factory,
                column_to_scope={column: "abc" for column in data_frame.columns.to_list()},
                domain="def",
                data_frame=data_frame,
                missing_property_columns=data_frame.columns.to_list(),
            )
        self.assertTrue(
            expected_message in str(context.exception), str(context.exception)
        )

    @parameterized.expand(
        [
            (
                "Invalid numeric type",
                pd.Series(data=[1], index=["a"]),
                pd.Series(data=["int16"], index=["a"]),
                "The following columns in the data_frame have not been mapped to LUSID data types: {'a': 'int16'}",
            )
        ]
    )
    def test_create_property_values(self, _, row, dtypes, expected_message) -> None:
        with self.assertRaises(TypeError) as context:
            cocoon.properties.create_property_values(
                row=row, scope="abc", domain="def", dtypes=dtypes, column_to_scope={column: "abc" for column in row}
            )
        self.assertTrue(
            expected_message in str(context.exception), str(context.exception)
        )



import re
import pandas as pd
from collections import defaultdict
from functools import reduce
from .record import Rec
from .either import Either
from . import back_compat

type_re = re.compile(r"(.*)\((.*),(.*)\)")


# Convert an iterable dataset to a DataFrame
def to_df(data, columns):
    if isinstance(data, Rec):
        data = data.content

    # Create record by accessing each column
    def to_record(obj):
        return {col: access(obj, col) for col in columns}

    def property_value(p):
        return p.metric_value.value if p.label_value is None else p.label_value

    # traverse dot notation to flatten sub-objects
    def access(obj, col):
        if col.startswith("P:"):
            try:
                props = getattr(obj[1], "properties")
                if type(props) == list:
                    return property_value(
                        [p.value for p in props if p.key == col[2:]][0]
                    )
                return property_value(props[col[2:]].value)
            except:
                return None

        if col.startswith("SHK:"):
            try:
                props = getattr(obj[1], "sub_holding_keys")
                return property_value(props[col[4:]].value)
            except:
                return None

        for fld in col.split("."):
            if fld.startswith("KEY:"):
                obj = obj.get(fld[4:]) if obj else None
            else:
                obj = getattr(obj, fld) if obj else None
        return obj

    # Try standard representations
    try:
        iterator = iter(data)
        if ( 'BaseModel.__iter' in str(iterator)):
            raise Exception
    except:
        iterator = iter(data.values)

    records = [to_record(o) for o in iterator]

    if len(records) == 0:
        return pd.DataFrame({col: [] for col in columns})[columns]

    return pd.DataFrame.from_records(records)[columns]


# Utilities to convert YYYY-MM-DD strings to and from UTC dates.
def to_date(date, **kwargs):
    return pd.to_datetime(date, utc=True, **kwargs) if date is not None else None


def from_date(date):
    return date.strftime("%Y-%m-%d") if date else None


def add_days(date, days):
    return to_date(date) + pd.Timedelta("{} days".format(days))


# Display a dataframe with no cropping
def display_df(df, decimals=2):
    fmt = "{:,." + str(decimals) + "f}"
    pd.options.display.float_format = fmt.format
    pd.set_option("max_colwidth", None)

    try:
        if len(df) == 1 and len(df.columns) > 5:
            df = df.T
        with pd.option_context("display.width", None, "display.max_rows", 1000):
            print(df.fillna(""))
    except:
        print(df)


# Create API objects from a dataframe
def from_df(
    df, record_type, complex_types, related=None, columns=None, date_fields=None
):
    simple_columns = []
    complex_columns = defaultdict(list)
    properties = []

    date_fields = set(
        (date_fields or [])
        + [k for (k, v) in record_type.openapi_types.items() if v == "datetime"]
        + [c for c in df.columns if pd.api.types.is_datetime64_any_dtype(df[c])]
    )

    if len(date_fields) > 0:
        df = df.copy()
        for col in date_fields:
            df[col] = pd.to_datetime(df[col], utc=True).map(
                lambda x: None if pd.isna(x) else x
            )

    for col in columns or df.columns.values:
        if col.startswith("P:"):
            properties.append(col[2:])
        elif "." in col:
            (k, v) = col.split(".")
            complex_columns[k].append(v)
        else:
            simple_columns.append(col)

    def build_complex_type(row, col, fields):
        d = {f: row["{}.{}".format(col, f)] for f in fields}
        col_type = complex_types[record_type.openapi_types[col]]
        return col_type(**d)

    def build_properties(row):
        ptype_tpl = type_re.findall(record_type.openapi_types["properties"])[0]

        ptype = complex_types[ptype_tpl[2].strip()]

        def prop_builder(property_key, value):
            if isinstance(value, str):
                return ptype(
                    key=property_key,
                    value=complex_types["PropertyValue"](label_value=value),
                )
            elif pd.isna(value) == False:
                return ptype(
                    key=property_key,
                    value=complex_types["PropertyValue"](
                        metric_value=complex_types["MetricValue"](value)
                    ),
                )
            else:
                return None

        props = [
            (col, prop_builder(col, row["P:{}".format(col)])) for col in properties
        ]
        d = dict([p for p in props if p[1] != None])
        return d

    def to_type(i, row):
        fields = {col: row[col] for col in simple_columns}
        fields.update(
            {
                col: build_complex_type(row, col, fields)
                for col, fields in complex_columns.items()
            }
        )

        if "properties" in record_type.openapi_types and len(properties) > 0:
            fields["properties"] = build_properties(row)

        if related != None:
            if callable(related):
                fields = related(i, row, fields)
            else:
                # Dict type
                fields.update(related.get(i, {}))

        # Quick and dirty instrument_uid handling
        # This should change to allow the full
        # instrument resolution logic to apply
        if "instrument_uid" in fields.keys():
            fields["instrument_identifiers"] = to_instrument_identifiers(
                fields["instrument_uid"]
            )
            del fields["instrument_uid"]

        # Remove any 'noise' from the dataframe
        allowed = set(record_type.openapi_types.keys())
        trimmed = dict([tpl for tpl in fields.items() if tpl[0] in allowed])

        return record_type(**trimmed)

    return [to_type(i, row) for (i, row) in df.iterrows()]


def to_instrument_identifiers(uid):
    if uid.startswith("CCY_"):
        return {"Instrument/default/Currency": uid[4:]}
    if uid.startswith("Ccy:"):
        return {"Instrument/default/Currency": uid[4:]}
    if uid.startswith("Currency:"):
        return {"Instrument/default/Currency": uid[9:]}
    elif uid.startswith("ClientInternal:"):
        return {"Instrument/default/ClientInternal": uid[15:]}
    elif uid.startswith("Figi:"):
        return {"Instrument/default/Figi": uid[5:]}
    elif uid.startswith("RIC:"):
        return {"Instrument/default/RIC": uid[4:]}
    else:
        parts = uid.split(":")
        if len(parts) == 2 and parts[0].startswith("Instrument"):
            return {parts[0]: parts[1]}
        return {"Instrument/default/LusidInstrumentId": uid}


# Convert iterable of Record to DataFrame
def records_to_df(records):
    return pd.DataFrame([r.to_dict() for r in records])


# Break a dataframe down into batches
def chunk(seq, size):
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


# Write statistics out
def dump_stats(filename, stats, columns):
    if len(stats) > 0:
        df = records_to_df(stats)[columns]
        if filename == "-":
            display_df(df.drop(["startTime", "endTime"], axis=1))
        else:
            df.to_csv(filename, index=False)


# Limit the length of DataFrame
def trim_df(df, limit, **kwargs):
    if kwargs.get("sort", None) != None:
        df = df.sort_values(kwargs["sort"]).reset_index(drop=True)
    return df[:limit] if 0 < limit < len(df) else df


# Template 'program'
def standard_flow(parser, connector, executor, display_df=display_df):
    args = parser()
    api = connector(args)

    either = Either(executor(api, args))

    # called if the executor returns a success
    def success(df):
        # Query type programs will return a dataframe
        if df is not None:
            fn = args.__dict__.get("filename", None)
            if fn is not None:
                if ".xls" in fn.lower():
                    df.to_excel(fn, index=False)
                elif fn.endswith(".pk"):
                    df.to_pickle(fn)
                else:
                    df.to_csv(fn, index=False)
            else:
                if "dfq" in args and args.dfq:
                    from . import dfq

                    dfq.dfq(dfq.parse(False, args.dfq), df)
                else:
                    return display_df(df)

    rv = either.match(left=display_error, right=success)

    api.dump_stats()

    return rv


# Nicely display an error from LUSID
def display_error(error, response=False):
    try:
        print(
            "ERROR: {} Reason:{}, Code:{}\n".format(
                error.status, error.reason, error.code
            )
        )
        print("MESSAGE: {}\n".format(error.message))
        print("DETAILS: {}\n".format(error.detailed_message))
        print("Instance: {}\n".format(error.instance))

        if len(error.items) > 0:
            df = records_to_df(
                [Rec(Id=key, Detail=item) for key, item in error.items.items()]
            )
            print("ITEMS (max 50)")
            display_df(df[:50])
    except:
        print(str(error))
    return response


# backwards compatibility
def read_csv(path, frame_type=None, **kwargs):
    return read_input(path, frame_type, **kwargs)


# Read in a data-file and apply any backwards compatibility settings
def read_input(path, frame_type=None, mappings=None, **kwargs):
    sheet = kwargs.get("sheet_name", 0)
    if is_path_supported_excel_with_sheet(path):
        path, sheet = path.rsplit(":", 1)

    if ".xls" in path.lower():
        df = pd.read_excel(path, sheet_name=sheet, engine="openpyxl", **kwargs)
    else:
        df = pd.read_csv(path, **kwargs)

    if mappings is not None:
        df = df.rename(columns=mappings)
        df = df[list(set(mappings.values()) & set(df.columns))]

    return back_compat.convert(frame_type, df)


# Check if a path is a supported excel file with a suffixed sheet
def is_path_supported_excel_with_sheet(path):
    return re.match(".*\.(xls|xlsx|xlsm|xlsb):", path)


# Create PerpetualProperties request from prefixed columns in a dataframe
def perpetual_upsert(models, df, prefix="P:"):
    offset = len(prefix)

    def make_property(properties, key):
        return properties + [
            (
                key[offset:],
                models.PerpetualProperty(
                    key[offset:], models.PropertyValue(label_value=str(value))
                ),
            )
            for value in df[key].dropna().head(1).tolist()
        ]

    return dict(
        reduce(
            make_property, [c for c in df.columns.values if c.startswith(prefix)], []
        )
    )


# Return the serialised representation of the object
def Serialise(api, body, bodyType):
    return api.api._serialize.body(body, bodyType)


def process_input(aliases, api, args, fn):
    df = pd.concat(
        [read_input(input_file, dtype=str) for input_file in args.input],
        ignore_index=True,
        sort=False,
    )
    if args.mappings:
        df.rename(
            columns=dict(
                [
                    (s[1], aliases.get(s[0], s[0]))
                    for s in [m.split("=") for m in args.mappings]
                ]
            ),
            inplace=True,
        )
    prop_keys = [col for col in df.columns.values if col.startswith("P:")]
    identifiers = [col for col in df.columns.values if col in args.identifiers]
    # Identifiers have to be unique
    df = df.drop_duplicates(identifiers)

    return fn(api, args, df, identifiers, prop_keys)



import logging
from pathlib import Path
import lusid
import lusid.models as models
import unittest
from finbourne_sdk_utils.pandas_utils.lusid_pandas import lusid_response_to_data_frame
import pandas as pd
import datetime
from finbourne_sdk_utils import logger
import os

class TestResponseToPandasObject(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))
        
        cls.api_factory = lusid.SyncApiClientFactory()

        cls.holding_no_props_1 = lusid.models.portfolio_holding.PortfolioHolding(
            cost={"amount": 549997.05, "currency": "GBP"},
            cost_portfolio_ccy={"amount": 0.0, "currency": "GBP"},
            holding_type="P",
            instrument_uid="LUID_XQ6VSO8F",
            properties={},
            settled_units=137088.0,
            sub_holding_keys={},
            transaction=None,
            units=137088.0,
        )

        cls.holding_no_props_2 = lusid.models.portfolio_holding.PortfolioHolding(
            cost={"amount": 12345.05, "currency": "GBP"},
            cost_portfolio_ccy={"amount": 0.0, "currency": "GBP"},
            holding_type="P",
            instrument_uid="LUID_123",
            properties={},
            settled_units=1372222.0,
            sub_holding_keys={},
            transaction=None,
            units=1372228.0,
        )

        cls.holding_1 = lusid.models.portfolio_holding.PortfolioHolding(
            cost={"amount": 12345.05, "currency": "GBP"},
            cost_portfolio_ccy={"amount": 0.0, "currency": "GBP"},
            holding_type="P",
            instrument_uid="LUID_123",
            properties={
                "Instrument/MultiAssetScope/shares_out": {
                    "effective_from": datetime.datetime(1, 1, 1, 0, 0),
                    "key": "Instrument/MultiAssetScope/shares_out",
                    "value": {
                        "label_value": None,
                        "metric_value": {"unit": None, "value": 100000.0},
                    },
                },
                "Instrument/MultiAssetScope/sm_instrument_type": {
                    "effective_from": datetime.datetime(1, 1, 1, 0, 0),
                    "key": "Instrument/MultiAssetScope/sm_instrument_type",
                    "value": {"label_value": "Equity", "metric_value": None},
                },
            },
            settled_units=1372222.0,
            sub_holding_keys={
                "Transaction/MultiAssetScope/PortionClass": {
                    "key": "Transaction/MultiAssetScope/PortionClass",
                    "value": {
                        "label_value": "Absolute " "Return " "(other)",
                        "metric_value": None,
                    },
                },
                "Transaction/MultiAssetScope/PortionRegion": {
                    "key": "Transaction/MultiAssetScope/PortionRegion",
                    "value": {"label_value": "None", "metric_value": None},
                },
                "Transaction/MultiAssetScope/PortionSubClass": {
                    "key": "Transaction/MultiAssetScope/PortionSubClass",
                    "value": {
                        "label_value": "Absolute " "Return " "(other)",
                        "metric_value": None,
                    },
                },
            },
            transaction=None,
            units=1372228.0,
        )

        cls.test_instrument_response_1 = lusid.models.instrument.Instrument(
            lusid_instrument_id="LUID_TEST1",
            version=1,
            name="Test LUID1",
            identifiers={
                "ClientInternal": "imd_35349922",
                "Figi": "BBG000BRVH22",
                "Isin": "GB00B1KJJ422",
                "LusidInstrumentId": "LUID_ZJJ1IY22",
                "Ticker": "WTB22",
            },
            state="Active",
        )

        cls.test_instrument_response_2 = lusid.models.instrument.Instrument(
            lusid_instrument_id="LUID_TEST2",
            version=1,
            name="Test LUID2",
            identifiers={
                "ClientInternal": "imd_35349900",
                "Figi": "BBG000BRVH05",
                "Isin": "GB00B1KJJ408",
                "LusidInstrumentId": "LUID_ZJJ1IYJS",
                "Ticker": "WTB",
            },
            state="Active",
        )

    def test_response_to_df(self):

        test_holdings_response = lusid.models.versioned_resource_list_of_portfolio_holding.VersionedResourceListOfPortfolioHolding(
            version=1, values=[self.holding_no_props_1, self.holding_no_props_2]
        )

        holdings_df = lusid_response_to_data_frame(test_holdings_response)
        self.assertEqual(type(holdings_df), pd.DataFrame)
        self.assertEqual(holdings_df.loc[0]["instrument_uid"], "LUID_XQ6VSO8F")

    def test_instrument_response_to_df(self):

        instrument_df = lusid_response_to_data_frame(self.test_instrument_response_1)
        self.assertEqual(type(instrument_df), pd.DataFrame)
        self.assertEqual(instrument_df.loc["lusid_instrument_id"][0], "LUID_TEST1")

    def test_transaction_alias_response_to_df(self):

        transaction_alias_response = lusid.models.transaction_configuration_type_alias.TransactionConfigurationTypeAlias(
            type="Buy",
            description="Purchase",
            transaction_class="Basic",
            transaction_group="Default",
            transaction_roles="LongLonger",
        )

        transaction_alias_df = lusid_response_to_data_frame(transaction_alias_response)
        transaction_alias_df_format = lusid_response_to_data_frame(
            transaction_alias_response, rename_properties=True
        )

        self.assertEqual(type(transaction_alias_df), pd.DataFrame)
        self.assertEqual(transaction_alias_df.loc["transaction_class"][0], "Basic")
        self.assertEqual(type(transaction_alias_df_format), pd.DataFrame)
        self.assertEqual(
            transaction_alias_df_format.loc["transaction_class"][0], "Basic"
        )

    def test_response_to_df_fail(self):

        self.assertRaises(
            TypeError, lambda: lusid_response_to_data_frame("test_string")
        )

    def test_response_none(self):

        self.assertRaises(TypeError, lambda: lusid_response_to_data_frame(None))

    def test_response_to_df_format(self):

        test_holdings_response = lusid.models.versioned_resource_list_of_portfolio_holding.VersionedResourceListOfPortfolioHolding(
            version=1, values=[self.holding_1]
        )

        holdings_df = lusid_response_to_data_frame(
            test_holdings_response, rename_properties=True
        )

        df_columns = holdings_df.columns

        self.assertEqual(type(holdings_df), pd.DataFrame)
        self.assertEqual(holdings_df.loc[0]["instrument_uid"], "LUID_123")
        self.assertIn("PortionClass(MultiAssetScope-SubHoldingKeys)", df_columns)
        self.assertIn("shares_out(MultiAssetScope-Properties)", df_columns)
        self.assertNotIn("JunkBadColumn(SubHoldingKey)", df_columns)

    def test_response_to_df_rename_mapping_format(self):

        rename_mapping = {
            "instrument_uid": "Instrument ID",
            "holding_type": "Holding Type",
            "units": "Units",
            "settled_units": "Settled Units",
            "cost.amount": "Quantity",
            "cost.currency": "Cost Currency",
        }

        test_holdings_response = lusid.models.versioned_resource_list_of_portfolio_holding.VersionedResourceListOfPortfolioHolding(
            version=1, values=[self.holding_1]
        )

        holdings_df = lusid_response_to_data_frame(
            test_holdings_response,
            rename_properties=True,
            column_name_mapping=rename_mapping,
        )

        df_columns = holdings_df.columns

        self.assertEqual(type(holdings_df), pd.DataFrame)
        self.assertEqual(holdings_df.loc[0]["Instrument ID"], "LUID_123")
        self.assertIn("PortionClass(MultiAssetScope-SubHoldingKeys)", df_columns)
        self.assertTrue(
            set(
                ["Holding Type", "Units", "Settled Units", "Quantity", "Cost Currency"]
            ).issubset(set(df_columns))
        )

    def test_response_to_df_rename_mapping_no_format(self):

        rename_mapping = {
            "instrument_uid": "Instrument ID",
            "holding_type": "Holding Type",
            "units": "Units",
            "settled_units": "Settled Units",
            "cost.amount": "Quantity",
            "cost.currency": "Cost Currency",
        }

        test_holdings_response = lusid.models.versioned_resource_list_of_portfolio_holding.VersionedResourceListOfPortfolioHolding(
            version=1, values=[self.holding_1]
        )

        holdings_df = lusid_response_to_data_frame(
            test_holdings_response,
            rename_properties=False,
            column_name_mapping=rename_mapping,
        )

        df_columns = holdings_df.columns

        self.assertEqual(type(holdings_df), pd.DataFrame)
        self.assertEqual(holdings_df.loc[0]["Instrument ID"], "LUID_123")
        self.assertEqual(
            set(
                ["Holding Type", "Units", "Settled Units", "Quantity", "Cost Currency"]
            ),
            set(
                ["Holding Type", "Units", "Settled Units", "Quantity", "Cost Currency"]
            ).intersection(set(df_columns)),
        )

    def test_response_to_df_rename_mapping_with_column_not_exist(self):

        rename_mapping = {
            "instrument_uid": "Instrument ID",
            "holding_type": "Holding Type",
            "test_not_exist": "Units",
            "settled_units": "Settled Units",
            "cost.amount": "Quantity",
            "cost.currency": "Cost Currency",
        }

        test_holdings_response = lusid.models.versioned_resource_list_of_portfolio_holding.VersionedResourceListOfPortfolioHolding(
            version=1, values=[self.holding_1]
        )

        holdings_df = lusid_response_to_data_frame(
            test_holdings_response,
            rename_properties=False,
            column_name_mapping=rename_mapping,
        )

        df_columns = holdings_df.columns

        self.assertEqual(type(holdings_df), pd.DataFrame)
        self.assertEqual(holdings_df.loc[0]["Instrument ID"], "LUID_123")
        self.assertEqual(
            set(["Holding Type", "Quantity", "Cost Currency"]),
            set(["Holding Type", "Quantity", "Cost Currency"]).intersection(
                set(df_columns)
            ),
        )
        self.assertNotIn("settled_units", df_columns)

    def test_response_to_df_rename_property(self):

        rename_mapping = {
            "instrument_uid": "Instrument ID",
            "holding_type": "Holding Type",
            "units": "Units",
            "settled_units": "Settled Units",
            "cost.amount": "Quantity",
            "cost.currency": "Cost Currency",
            "properties.Instrument/MultiAssetScope/shares_out.key": "Shares Outstanding",
        }

        test_holdings_response = lusid.models.versioned_resource_list_of_portfolio_holding.VersionedResourceListOfPortfolioHolding(
            version=1, values=[self.holding_1]
        )

        holdings_df = lusid_response_to_data_frame(
            test_holdings_response,
            rename_properties=False,
            column_name_mapping=rename_mapping,
        )

        df_columns = holdings_df.columns

        self.assertEqual(type(holdings_df), pd.DataFrame)
        self.assertEqual(holdings_df.loc[0]["Instrument ID"], "LUID_123")
        self.assertEqual(
            set(
                [
                    "Holding Type",
                    "Units",
                    "Quantity",
                    "Cost Currency",
                    "Shares Outstanding",
                ]
            ),
            set(
                [
                    "Holding Type",
                    "Units",
                    "Quantity",
                    "Cost Currency",
                    "Shares Outstanding",
                ]
            ).intersection(set(df_columns)),
        )

    def test_list_of_lusid_objects(self):

        list_response = [
            self.test_instrument_response_1,
            self.test_instrument_response_2,
        ]

        instrument_df = lusid_response_to_data_frame(list_response)
        self.assertEqual(type(instrument_df), pd.DataFrame)
        self.assertEqual(
            instrument_df.columns.to_list(),
            [
                "lusid_instrument_id",
                "version",
                "name",
                "identifiers.ClientInternal",
                "identifiers.Figi",
                "identifiers.Isin",
                "identifiers.LusidInstrumentId",
                "identifiers.Ticker",
                "state",
            ],
        )
        self.assertEqual(
            instrument_df["lusid_instrument_id"].to_list()[0], "LUID_TEST1"
        )

    def test_list_of_different_lusid_objects(self):

        with self.assertRaises(TypeError) as error:

            list_response = [
                self.test_instrument_response_1,
                self.test_instrument_response_2,
                self.holding_1,
            ]

            lusid_response_to_data_frame(list_response)

        self.assertEqual(
            error.exception.args[0], "All items in list must be of the same data type",
        )

    def test_list_of_same_lusid_objects_no_to_dict(self):

        with self.assertRaises(TypeError) as error:

            bad_list = ["random_string_1", "random_string_2"]

            lusid_response_to_data_frame(bad_list)

        self.assertEqual(
            error.exception.args[0],
            "All object items in list must have a to_dict() method",
        )

    def test_empty_list_lusid_response(self):

        empty_list = []

        empty_df = lusid_response_to_data_frame(empty_list)

        logging.info(empty_df)

        self.assertEqual(empty_df.empty, True)



import os
import unittest
from datetime import datetime, timedelta
from pathlib import Path
import pytz
import copy

import lusid
import finbourne_access
import finbourne_identity
from finbourne_access import models as access_models
from finbourne_identity import models as identity_models

from finbourne_sdk_utils import iam as iam
from finbourne_sdk_utils import logger

@unittest.skip('To be depricated')
class IAMTestsRoles(unittest.TestCase):
    api_factory = None
    access_api_factory = None
    identity_api_factory = None

    @classmethod
    def setUpClass(cls) -> None:
        
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

        api_secrets_filename = Path(__file__).parent.parent.parent.joinpath("secrets.json")
 
        # Get the LUSID API URL and the access token from the current api_factory
      
        secretsConfig = lusid.extensions.SecretsFileConfigurationLoader(api_secrets_file = api_secrets_filename)
        
        settings = secretsConfig.load_config()
    
        def build_api_url(base_url, endpoint):
            return f"{base_url[0: base_url.rfind('/')] }/{endpoint}"

        access_api_url = build_api_url(settings["api_url"], "access")
        identity_api_url = build_api_url(settings["api_url"], "identity")
        
        accessSettings = copy.copy(settings)
        identifySettings = copy.copy(settings)
        
        accessSettings['api_url'] = access_api_url
        identifySettings['api_url'] = identity_api_url

        cls.access_api_factory = finbourne_access.extensions.SyncApiClientFactory( config_loaders =  [lusid.extensions.ArgsConfigurationLoader(**accessSettings)])
        
        cls.identity_api_factory = finbourne_identity.extensions.SyncApiClientFactory( config_loaders = [lusid.extensions.ArgsConfigurationLoader(**identifySettings)])


    def test_create_role_success(self) -> None:
        """
        Test that a role can be created successfully

        :return: None
        """

        role_code = "TestRoleCode"

        try:
            self.delete_role(role_code)
        except Exception as e:
            print(e)

        # Create a role creation request (using the access model)
        access_role_creation_request = create_access_role_creation_request(role_code)

        # Create a role using the LPT create_role method
        iam.roles.create_role(self.access_api_factory, self.identity_api_factory, access_role_creation_request)


        # Check that the role was correctly created through the access API
        access_role = self.get_access_role(role_code)
        self.assertEqual(first=access_role.id.code, second=role_code)

        # Check that the role was correctly created through the identity API
        identity_role = self.get_identity_role(role_code)
        self.assertEqual(first=identity_role.role_id.code, second=role_code)

    def delete_role(self, role_code):
        access_roles_api = self.access_api_factory.build(finbourne_access.RolesApi)
        identity_roles_api = self.identity_api_factory.build(
            finbourne_identity.RolesApi
        )

        # Try to delete the role through the access API
        try:
            access_roles_api.delete_role(role_code)
        except Exception as e:
            print(e)

        identity_role = self.get_identity_role(role_code)
        try:
            identity_roles_api.delete_role(identity_role.id)
        except Exception as e:
            print(e)

    def get_access_role(self, role_code):
        access_roles_api = self.access_api_factory.build(finbourne_access.RolesApi)

        return access_roles_api.get_role(role_code)

    def get_identity_role(self, role_code):
        # Build the identity API factory
        identity_roles_api = self.identity_api_factory.build(
            finbourne_identity.RolesApi
        )

        roles = identity_roles_api.list_roles()

        return next((r for r in roles if r.role_id.code == role_code))


def create_access_role_creation_request(role_code):
    return access_models.RoleCreationRequest(
        code=role_code,
        description="Test Role Description",
        resource=access_models.RoleResourceRequest(
            policy_id_role_resource=access_models.PolicyIdRoleResource(
                policies=[
                    access_models.PolicyId(scope="default", code="TestPolicyCode")
                ]
            )
        ),
        when=access_models.WhenSpec(
            activate=datetime.now(tz=pytz.utc) - timedelta(days=2),
            deactivate=datetime(9999, 12, 31, tzinfo=pytz.utc),
        ),
    )



import os
import re
import unittest
import json
from pathlib import Path
import pandas as pd
import lusid
from finbourne_sdk_utils import logger
from finbourne_sdk_utils.cocoon.cocoon import load_from_data_frame
from finbourne_sdk_utils.cocoon.cocoon_printer import format_portfolios_response


class CocoonPrinterIntegrationTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        
        cls.api_factory = lusid.SyncApiClientFactory()
        
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    def test_format_portfolios_response_includes_extended_errors(self):
        """
        Test that the cocoon printer also returns the RequestID and ErrorDetails for any errors in the response.
        """
        # Load a portfolio that will contain an error using load_from_data_frame
        data_frame = pd.read_csv(
            Path(__file__).parent.joinpath("data/cocoon_format_portfolio_errors.csv")
        )
        portfolio_response = load_from_data_frame(
            api_factory=self.api_factory,
            scope="cocoon_format_errors",
            data_frame=data_frame,
            mapping_required={
                "code": "FundCode",
                "display_name": "display_name",
                "created": "created",
                "base_currency": "base_currency",
            },
            mapping_optional={"description": "description", "accounting_method": None},
            file_type="portfolios",
        )

        # Pass the response into the cocoon printer formatter
        succ, err = format_portfolios_response(
            portfolio_response, extended_error_details=True
        )

        # Assert that the error part includes RequestID details and ErrorDetails
        self.assertEqual(2, len(succ))
        self.assertEqual(1, len(err))
        for index, row in err.iterrows():
            self.assertEqual(row[err.columns[0]], "Bad Request")
            
            
            # regex for new and old request id patterns
            patterns = [r"([A-Z0-9]+:[A-Z0-9]+)", r"(\d{10}-[a-f0-9]{32})"]

            foundRequestId = False
            for pattern in patterns:
                match = re.search(pattern, row[err.columns[2]])
                if match:
                    print(f"{match.group(1)} matched pattern: {pattern}")
                    foundRequestId = True
                    break
                
            self.assertTrue(foundRequestId, "No request id found in error message")
                    
            # Deserialise the ErrorDetails field and check one of the values
            self.assertEqual(
                json.loads(row[err.columns[3]]).get("name"), "UndefinedCurrencyFailure"
            )

        # Delete the portfolios at the end of the test
        for portfolio in portfolio_response.get("portfolios").get("success"):
            self.api_factory.build(lusid.api.PortfoliosApi).delete_portfolio(
                scope="cocoon_format_errors", code=portfolio.id.code
            )


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest import mock
import os
from finbourne_sdk_utils.lpt import connect_lusid


class ConnectLusidTests(unittest.TestCase):
    @mock.patch.dict(os.environ, {"FBN_LUSID_API_URL": ""})
    def test_missing_one_config(self):
        config = {}
        config["api"] = {}
        config["api"]["tokenUrl"] = "abc"
        config["api"]["username"] = "abc"
        config["api"]["password"] = os.environ.get("FBN_PASSWORD", "dummyPassword")
        config["api"]["clientId"] = "abc"
        config["api"]["clientSecret"] = "abc"
        config["api"]["apiUrl"] = None

        missing_config = connect_lusid.check_for_missing_config(config)

        self.assertEqual(1, len(missing_config))
        self.assertEqual(
            [{"Env variable": "FBN_LUSID_API_URL", "Secrets file key": "apiUrl"}],
            missing_config,
        )



import os
import unittest
import uuid
from pathlib import Path

import pandas
import pandas as pd
import lusid

from finbourne_sdk_utils import cocoon as cocoon
from parameterized import parameterized
from finbourne_sdk_utils import logger
from finbourne_sdk_utils.cocoon.utilities import create_scope_id


class CocoonTestsPortfolios(unittest.TestCase):
    api_factory = None

    @classmethod
    def setUpClass(cls) -> None:
        
        cls.api_factory = lusid.SyncApiClientFactory()
        
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    @parameterized.expand(
        [
            [
                "Standard load",
                "prime_broker_test",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {"description": "description", "accounting_method": None},
                {},
                ["base_currency"],
                "operations001",
                None,
                None,
                None,
            ],
            [
                "Standard load with ~700 portfolios",
                "prime_broker_test",
                "data/metamorph_portfolios-large.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {"description": "description", "accounting_method": None},
                {},
                ["base_currency"],
                "operations001",
                None,
                None,
                None,
            ],
            [
                "Add in some constants",
                "prime_broker_test",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {"description": "description", "accounting_method": "$AverageCost"},
                {},
                ["base_currency"],
                "operations001",
                None,
                None,
                None,
            ],
            [
                "Standard load with unique properties scope to ensure property creation",
                "prime_broker_test",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {"description": "description", "accounting_method": None},
                {},
                ["base_currency"],
                f"operations001_{create_scope_id()}",
                None,
                None,
                None,
            ],
            [
                "Standard load with a single sub-holding-key",
                f"prime_broker_test_{create_scope_id(use_uuid=True)}",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {"description": "description", "accounting_method": None},
                {},
                ["base_currency"],
                "operations001",
                ["Strategy"],
                None,
                ["Transaction/operations001/Strategy"],
            ],
            [
                "Standard load with a multiple sub-holding-keys in a unique scope",
                f"prime_broker_test_{create_scope_id(use_uuid=True)}",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {"description": "description", "accounting_method": None},
                {},
                ["base_currency"],
                f"operations001",
                ["Strategy", "Broker"],
                None,
                [
                    "Transaction/operations001/Strategy",
                    "Transaction/operations001/Broker",
                ],
            ],
            [
                "Standard load with a sub-holding-key with scope specified",
                f"prime_broker_test_{create_scope_id(use_uuid=True)}",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {"description": "description", "accounting_method": None},
                {},
                ["base_currency"],
                f"operations001_{create_scope_id(use_uuid=True)}",
                ["Trader/Strategy"],
                None,
                ["Transaction/Trader/Strategy"],
            ],
            [
                "Standard load with a sub-holding-key with domain and scope specified",
                f"prime_broker_test_{create_scope_id(use_uuid=True)}",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {"description": "description", "accounting_method": None},
                {},
                ["base_currency"],
                f"operations001_{create_scope_id(use_uuid=True)}",
                ["Transaction/Trader/Strategy"],
                None,
                ["Transaction/Trader/Strategy"],
            ],
            [
                "Standard load with a single sub-holding-key in a different scope",
                f"prime_broker_test_{create_scope_id(use_uuid=True)}",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {"description": "description", "accounting_method": None},
                {},
                ["base_currency"],
                "operations001",
                ["Strategy"],
                f"accountview007",
                ["Transaction/accountview007/Strategy"],
            ],
        ]
    )
    def test_load_from_data_frame_a_portfolios_success(
        self,
        _,
        scope,
        file_name,
        mapping_required,
        mapping_optional,
        identifier_mapping,
        property_columns,
        properties_scope,
        sub_holding_keys,
        sub_holdings_key_scope,
        expected_sub_holdings_keys,
    ) -> None:
        """
        Test that portfolios can be loaded successfully

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict(str, str) mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict(str, str) mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict(str, str) identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list(str) property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to
        :param str sub_holdings_key_scope: The scope of the sub holding keys
        :param list(str): Expected sub holding keys

        :return: None
        """
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

       
        responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="portfolios",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
            sub_holding_keys=sub_holding_keys,
            sub_holding_keys_scope=sub_holdings_key_scope,
        )
       
        self.assertEqual(
            first=len(responses["portfolios"]["success"]), second=len(data_frame)
        )

        response_codes = [
            response.id.code
            if isinstance(response, lusid.models.Portfolio)
            else response.origin_portfolio_id.code
            for response in responses["portfolios"]["success"]
        ]

        self.assertEqual(
            first=response_codes,
            second=list(data_frame[mapping_required["code"]].values),
        )

        # Assert that by no unmatched_identifiers are returned in the response for portfolios
        self.assertFalse(responses["portfolios"].get("unmatched_identifiers", False))

        if expected_sub_holdings_keys is None:
            return

        for portfolio_code in response_codes:

            portfolio_details = self.api_factory.build(
                lusid.api.TransactionPortfoliosApi
            ).get_details(scope=scope, code=portfolio_code)

            self.assertSetEqual(
                set(portfolio_details.sub_holding_keys), set(expected_sub_holdings_keys)
            )
        
    @parameterized.expand(
        [
            [
                "ApiException due to invalid scope",
                "p rime_broker_test", # not the space between p and r, this should cause a server side response error
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {"description": "description", "accounting_method": None},
                {},
                ["base_currency"],
                "operations001",
                None,
            ],
        ]
    )
    def test_load_from_data_frame_a_portfolios_failure(
        self,
        _,
        scope,
        file_name,
        mapping_required,
        mapping_optional,
        identifier_mapping,
        property_columns,
        properties_scope,
        sub_holding_keys,
    ) -> None:
        """
        Test that portfolios can be loaded successfully

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list[str] property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to
        :param any expected_outcome: The expected outcome

        :return: None
        """
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))
       
        file_type = "portfolios"
        
        responses = (
            cocoon.cocoon.load_from_data_frame(
                api_factory=self.api_factory,
                scope=scope,
                data_frame=data_frame,
                mapping_required=mapping_required,
                mapping_optional=mapping_optional,
                file_type=file_type,
                identifier_mapping=identifier_mapping,
                property_columns=property_columns,
                properties_scope=properties_scope,
                sub_holding_keys=sub_holding_keys,
            ),
        )
        self.assertEqual( len(responses) , 1 )
        self.assertEqual( len(responses[0][file_type]['errors']), 3)
        
        self.assertTrue( isinstance(responses[0][file_type]['errors'][0], lusid.exceptions.ApiException ))
        self.assertTrue( isinstance(responses[0][file_type]['errors'][1], lusid.exceptions.ApiException ))
        self.assertTrue( isinstance(responses[0][file_type]['errors'][2], lusid.exceptions.ApiException ))
        
        # body is a str, not a type
        # avoided a hard coded message check as this likely to break wen error message is changed
        self.assertTrue( "scope" in responses[0][file_type]['errors'][0].body)
        self.assertTrue( "scope" in responses[0][file_type]['errors'][1].body)
        self.assertTrue( "scope" in responses[0][file_type]['errors'][2].body)

    @parameterized.expand(
        [
            [
                "Standard load",
                "prime_broker_test",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {},
                {},
                ["base_currency"],
                "operations0011",
            ]
        ]
    )
    def test_properties_list_strings(
        self,
        _,
        scope,
        file_name,
        mapping_required,
        mapping_optional,
        identifier_mapping,
        property_columns,
        properties_scope,
    ) -> None:
        """
        Test that portfolios can be loaded successfully

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict(str, str) mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict(str, str) mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict(str, str) identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list(str) property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to

        :return: None
        """
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="portfolios",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
        )

        for property_column in property_columns:
            response = self.api_factory.build(
                lusid.api.PropertyDefinitionsApi
            ).get_property_definition(
                domain="Portfolio", scope=properties_scope, code=property_column,
            )
            self.assertEqual(
                f"Portfolio/{properties_scope}/{property_column}", response.key
            )

    @parameterized.expand(
        [
            [
                "Source only",
                f"prime_broker_test__{uuid.uuid4()}",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {},
                {},
                {"source": "base_currency"},
                "operations0011",
                "operations0011",
                "base_currency",
            ],
            [
                "Source and target",
                f"prime_broker_test_{uuid.uuid4()}",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {},
                {},
                {"source": "base_currency", "target": "base_currency2"},
                "operations0011",
                "operations0011",
                "base_currency2",
            ],
            [
                "Scope",
                f"prime_broker_test_{uuid.uuid4()}",
                "data/metamorph_portfolios-unique.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {},
                {},
                {"source": "base_currency", "target": "base_currency2", "scope": "foo"},
                "operations0011",
                "foo",
                "base_currency2",
            ],
        ]
    )
    def test_properties_dicts(
        self,
        _,
        scope,
        file_name,
        mapping_required,
        mapping_optional,
        identifier_mapping,
        property_column,
        properties_scope,
        expected_property_scope,
        expected_property_code,
    ) -> None:
        """
        Test that portfolios can be loaded successfully

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict(str, str) mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict(str, str) mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict(str, str) identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list(str) property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to

        :return: None
        """
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="portfolios",
            identifier_mapping=identifier_mapping,
            property_columns=[property_column],
            properties_scope=properties_scope,
        )

        response = self.api_factory.build(
            lusid.api.PropertyDefinitionsApi
        ).get_property_definition(
            domain="Portfolio",
            scope=expected_property_scope,
            code=expected_property_code,
        )
        self.assertEqual(
            f"Portfolio/{expected_property_scope}/{expected_property_code}",
            response.key,
        )

        for code in data_frame["FundCode"].values:
            response = self.api_factory.build(
                lusid.api.PortfoliosApi
            ).get_portfolio_properties(scope=scope, code=code)
            self.assertIsNotNone(
                response.properties.get(
                    f"Portfolio/{expected_property_scope}/{expected_property_code}"
                )
            )

    @parameterized.expand(
        [
            [
                "Invalid dictionary",
                [{"foo": "bar"}],
                "The value [{'foo': 'bar'}] provided in property_columns is invalid. "
                "{'foo': 'bar'} does not contain the mandatory 'source' key.",
            ],
            [
                "Non string or dictionary",
                [1],
                "The value [1] provided in property_columns is invalid. "
                "1 is not a string or dictionary.",
            ],
        ]
    )
    def test_invalid_properties(
        self, _, property_columns, expected_error_message
    ) -> None:
        with self.assertRaises(ValueError) as context:
            cocoon.cocoon.load_from_data_frame(
                api_factory=self.api_factory,
                scope="foo",
                data_frame=pandas.DataFrame(),
                mapping_required={},
                mapping_optional={},
                file_type="portfolios",
                identifier_mapping={},
                property_columns=property_columns,
                properties_scope="bar",
            )

        self.assertEqual(expected_error_message, str(context.exception))



import os
import unittest
from finbourne_sdk_utils import logger
from finbourne_sdk_utils.cocoon.validator import Validator
from parameterized import parameterized


class CocoonUtilitiesTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    @parameterized.expand(
        [
            [
                "Value does exist in allowed values",
                "Portfolio",
                "file_type",
                ["Portfolio", "Transaction"],
            ]
        ]
    )
    def test_check_allowed_value_success(self, _, value, value_name, allowed_values):

        Validator(value, value_name).check_allowed_value(allowed_values)

    @parameterized.expand(
        [
            [
                "Value does not exist in allowed values",
                "Reference",
                "file_type",
                ["Portfolio", "Transaction"],
                ValueError,
            ],
            [
                "Value does not exist in allowed values with empty list",
                "Reference",
                "file_type",
                [],
                ValueError,
            ],
            [
                "Value is not a string",
                1,
                "file_type",
                ["Portfolio", "Transaction"],
                ValueError,
            ],
            [
                "Value is a list",
                ["Portfolio", "Transaction"],
                "file_type",
                ["Portfolio", "Transaction"],
                ValueError,
            ],
            [
                "Value is a dictionary",
                {"type": "portfolio"},
                "file_type",
                ["Portfolio", "Transaction"],
                ValueError,
            ],
        ]
    )
    def test_check_allowed_value_exception(
        self, _, value, value_name, allowed_values, expected_exception
    ):

        with self.assertRaises(expected_exception):
            Validator(value, value_name).check_allowed_value(allowed_values)

    @parameterized.expand(
        [
            [
                "A plural with a single 's' on the end",
                "Transactions",
                "file_type",
                "Transaction",
            ],
            [
                "A plural with a two 's' on the end",
                "Transactionss",
                "file_type",
                "Transaction",
            ],
            ["Already singular", "Transaction", "file_type", "Transaction"],
            ["Not a string", 1, "file_type", 1],
        ]
    )
    def test_make_singular(self, _, value, value_name, expected_outcome):

        singular = Validator(value, value_name).make_singular()

        self.assertIsInstance(singular, Validator)

        self.assertEqual(first=singular.value, second=expected_outcome)

    @parameterized.expand(
        [
            ["All upper case", "TRANSACTIONS", "file_type", "transactions"],
            ["Mixed case", "TrAnSaCTIons", "file_type", "transactions"],
            ["Already lower case", "transactions", "file_type", "transactions"],
            ["Not a string", 1, "file_type", 1],
        ]
    )
    def test_make_lower(self, _, value, value_name, expected_outcome):

        lower_case = Validator(value, value_name).make_lower()

        self.assertIsInstance(lower_case, Validator)

        self.assertEqual(first=lower_case.value, second=expected_outcome)

    @parameterized.expand(
        [
            ["None value provided with default", None, "batch_size", 10, 10],
            [
                "None value provided with None as default",
                None,
                "batch_size",
                None,
                None,
            ],
            ["Not None provided with default", 3, "batch_size", 10, 3],
        ]
    )
    def test_set_default_value_if_none(
        self, _, value, value_name, default, expected_outcome
    ):

        updated_value = Validator(value, value_name).set_default_value_if_none(default)

        self.assertIsInstance(updated_value, Validator)

        self.assertEqual(first=updated_value.value, second=expected_outcome)

    @parameterized.expand(
        [
            ["Override Flag Set to True", 10, "batch_size", True, 20, 20],
            ["Override Flag Set to False", 10, "batch_size", False, 20, 10],
            [
                "Override Flag Set to FalseTrue via expression",
                10,
                "batch_size",
                "Portfolio" == "Transaction",
                20,
                10,
            ],
        ]
    )
    def test_override_value(
        self, _, value, value_name, override_flag, override_value, expected_outcome
    ):

        updated_value = Validator(value, value_name).override_value(
            override_flag, override_value
        )

        self.assertIsInstance(updated_value, Validator)

        self.assertEqual(first=updated_value.value, second=expected_outcome)

    @parameterized.expand(
        [
            [
                "Single dictionary with None values",
                {"type": None, "code": "portfolio_code"},
                "optional_mapping",
                {"code": "portfolio_code"},
            ],
            [
                "Nested dictionary with None values at the bottom",
                {
                    "type": {"column": None, "default": "Reference"},
                    "code": "portfolio_code",
                },
                "required_mapping",
                {
                    "type": {"column": None, "default": "Reference"},
                    "code": "portfolio_code",
                },
            ],
            [
                "Dictionary with no None values",
                {"type": "Buy", "code": "portfolio_code"},
                "optional_mapping",
                {"type": "Buy", "code": "portfolio_code"},
            ],
        ]
    )
    def test_discard_dict_keys_none_value(self, _, value, value_name, expected_outcome):

        update_dict = Validator(value, value_name).discard_dict_keys_none_value()

        self.assertIsInstance(update_dict, Validator)

        self.assertEqual(first=update_dict.value, second=expected_outcome)

    @parameterized.expand(
        [
            [
                "Single dictionary with None values",
                {"type": None, "code": "portfolio_code"},
                "optional_mapping",
                [None, "portfolio_code"],
            ],
            [
                "Nested dictionary with None values at the bottom",
                {
                    "type": {"column": None, "default": "Reference"},
                    "code": "portfolio_code",
                },
                "required_mapping",
                [{"column": None, "default": "Reference"}, "portfolio_code"],
            ],
            [
                "Dictionary with no None values",
                {"type": "Buy", "code": "portfolio_code"},
                "optional_mapping",
                ["Buy", "portfolio_code"],
            ],
        ]
    )
    def test_get_dict_values(self, _, value, value_name, expected_outcome):

        dict_values = Validator(value, value_name).get_dict_values()

        self.assertIsInstance(dict_values, Validator)

        self.assertEqual(first=dict_values.value, second=expected_outcome)

    @parameterized.expand(
        [
            [
                "List with some members containing first character",
                ["$Code", "$Buy", "Sell"],
                "code_list",
                "$",
                ["Sell"],
            ],
            [
                "List with non members containing first character",
                ["Code", "Buy", "Sell"],
                "code_list",
                "$",
                ["Code", "Buy", "Sell"],
            ],
            [
                "First character has more than one character",
                ["$Code", "$Buy", "Sell"],
                "code_list",
                "$$",
                ["$Code", "$Buy", "Sell"],
            ],
        ]
    )
    def test_filter_list_using_first_character(
        self, _, value, value_name, first_character, expected_outcome
    ):

        updated_list = Validator(value, value_name).filter_list_using_first_character(
            first_character
        )

        self.assertIsInstance(updated_list, Validator)

        self.assertEqual(first=updated_list.value, second=expected_outcome)

    @parameterized.expand(
        [
            [
                "List is a subset",
                ["Portfolio", "Transaction"],
                "file_types",
                ["Portfolio", "Transaction", "Holding"],
                "all_file_types",
            ]
        ]
    )
    def test_check_subset_of_list_success(
        self, _, value, value_name, superset, superset_name
    ):

        Validator(value, value_name).check_subset_of_list(superset, superset_name)

    @parameterized.expand(
        [
            [
                "List is not a subset",
                ["Portfolio", "Transaction", "Quote"],
                "file_types",
                ["Portfolio", "Transaction", "Holding"],
                "all_file_types",
                ValueError,
            ]
        ]
    )
    def test_check_subset_of_list_exception(
        self, _, value, value_name, superset, superset_name, expected_exception
    ):

        with self.assertRaises(expected_exception):
            Validator(value, value_name).check_subset_of_list(superset, superset_name)

    @parameterized.expand(
        [
            [
                "Lists have no intersection",
                ["Portfolio", "Transaction", "Quote"],
                "file_types",
                ["Instrument", "Holding"],
                "other_file_types",
            ]
        ]
    )
    def test_check_no_intersection_with_list_success(
        self, _, value, value_name, other_list, list_name
    ):

        Validator(value, value_name).check_no_intersection_with_list(
            other_list, list_name
        )

    @parameterized.expand(
        [
            [
                "Lists have at one intersection",
                ["Portfolio", "Transaction", "Quote"],
                "file_types",
                ["Instrument", "Holding", "Quote"],
                "other_file_types",
                ValueError,
            ]
        ]
    )
    def test_check_no_intersection_with_list_exception(
        self, _, value, value_name, other_list, list_name, expected_exception
    ):

        with self.assertRaises(expected_exception):
            Validator(value, value_name).check_no_intersection_with_list(
                other_list, list_name
            )

    @parameterized.expand(
        [
            [
                "Dictionary missing 'source' key",
                [{"foo": "bar"}],
                "The value [{'foo': 'bar'}] provided in property_columns is invalid.",
                "{'foo': 'bar'} does not contain the mandatory 'source' key.",
            ],
            [
                "Dictionary with 'source' that's not a string",
                [{"source": 2}],
                "The value [{'source': 2}] provided in property_columns is invalid.",
                "2 in {'source': 2} is not a string.",
            ],
            [
                "Non string",
                [1],
                "The value [1] provided in property_columns is invalid",
                "1 is not a string or dictionary.",
            ],
            [
                "Multiple errors",
                [1, {"foo": "bar"}, {"source": [5]}],
                "The value [1, {'foo': 'bar'}, {'source': [5]}] provided in property_columns is invalid",
                "1 is not a string or dictionary, "
                + "{'foo': 'bar'} does not contain the mandatory 'source' key, "
                + "[5] in {'source': [5]} is not a string.",
            ],
        ]
    )
    def test_check_entries_are_strings_or_dict_containing_key_invalid_values(
        self, _, value, expected_message_part1, expected_message_part2
    ):
        with self.assertRaises(ValueError) as context:
            Validator(
                value, "property_columns"
            ).check_entries_are_strings_or_dict_containing_key("source")

        self.assertTrue(
            expected_message_part1 in str(context.exception), str(context.exception)
        )
        self.assertTrue(
            expected_message_part2 in str(context.exception), str(context.exception)
        )

    @parameterized.expand(
        [
            ["Dictionary with 'source'", [{"source": "foo"}]],
            ["String", ["foo"]],
            ["Multiple values", ["foo", {"source": "bar"}]],
        ]
    )
    def test_check_entries_are_strings_or_dict_containing_key_valid_values(
        self, _, value
    ):
        Validator(
            value, "property_columns"
        ).check_entries_are_strings_or_dict_containing_key("source")



import pandas as pd
import dateutil
from finbourne_sdk_utils.lpt import lpt
from finbourne_sdk_utils.lpt import lse
from finbourne_sdk_utils.lpt import stdargs

TOOLNAME = "instr_search"
TOOLTIP = "Search for Instruments"


def parse(extend=None, args=None):
    return (
        stdargs.Parser("Search Instruments", ["filename", "limit"])
        .add("--properties", nargs="*", help="properties to search")
        .add("--date")
        .extend(extend)
        .parse(args)
    )


def process_args(api, args):
    def success(result):
        flat = [i for r in result.content for i in r.mastered_instruments]
        if len(flat) > 0:
            identifiers = sorted(set.union(*[set(i.identifiers.keys()) for i in flat]))
            df = lpt.to_df(
                flat,
                ["name"] + ["identifiers.KEY:" + i + ".value" for i in identifiers],
            )
            df.columns = ["Name"] + identifiers
            return df
        else:
            return "No Matches"

    request = [
        api.models.InstrumentSearchProperty(s[0], s[1])
        for s in [p.split("=") for p in args.properties]
    ]
    return api.call.instruments_search(
        instrument_search_property=request,
        mastered_only=True,
        mastered_effective_at=lpt.to_date(args.date),
    ).bind(success)


# Standalone tool
def main(parse=parse, display_df=lpt.display_df):
    lpt.standard_flow(parse, lse.connect, process_args, display_df)
